import numpy as np
from scipy import integrate

from nopapy.utils.check import check_input, check_estimate
from nopapy.utils.setting import DEFAULT_BANDWIDTH, DEFAULT_KERNEL, DEFAULT_DEGREE

__all__ = ["NWEstimate", "GMEstimate", "LPEstimate"]

@check_input
@check_estimate
def NWEstimate(x: np.array, y: np.array, x0, h=DEFAULT_BANDWIDTH, k=DEFAULT_KERNEL) -> float:
    """
    Nadaraya-Watson estimate.
    Estimate the target value at x0 using the kernel function k and bandwidth h.

    Parameters:
        x (np.array): Independent variable.
        y (np.array): Dependent variable.
        x0 (float): Point at which to estimate the target value.
        h (float, optional): Bandwidth parameter. Defaults to DEFAULT_BANDWIDTH.
        k (callable, optional): Kernel function. Defaults to DEFAULT_KERNEL.

    Returns:
        float: Estimated value at given point x0.
    """
    l = k((x0 - x) / h) / sum(k((x0 - x) / h))
    return sum(l * y)

@check_input
@check_estimate
def GMEstimate(x: np.array, y: np.array, x0, h=DEFAULT_BANDWIDTH, k=DEFAULT_KERNEL) -> float:
    """
    Gasser-Muller estimate.
    Estimate the target value at x0 using the kernel function k and bandwidth h.

    Parameters:
        x (np.array): Independent variable.
        y (np.array): Dependent variable.
        x0 (float): Point at which to estimate the target value.
        h (float, optional): Bandwidth parameter. Defaults to DEFAULT_BANDWIDTH.
        k (callable, optional): Kernel function. Defaults to DEFAULT_KERNEL.

    Returns:
        float: Estimated value at given point x0.
    """
    s = [-np.inf] + [0.5 * (x[i] + x[i + 1]) for i in range(len(x) - 1)] + [np.inf]
    return sum([integrate.quad(k, (s[i] - x0) / h, (s[i + 1] - x0) / h)[0] * y[i] for i in range(len(x))])




@check_input
@check_estimate
def LPEstimate(x: np.array, y: np.array, x0, h=DEFAULT_BANDWIDTH, k=DEFAULT_KERNEL, p=DEFAULT_DEGREE) -> float:
    """
    Local polynomial estimate.
    Estimate the target values at x0 using the kernel function k, bandwidth h and degree p.

    Args:
        x (np.array): Independent variable.
        y (np.array): Dependent variable.
        x0 (float): Point at which to estimate the target value.
        h (float, optional): Bandwidth parameter. Defaults to DEFAULT_BANDWIDTH.
        k (callable, optional): Kernel function. Defaults to DEFAULT_KERNEL.
        p (int, optional): The degree of the local polynomial regression. Defaults to DEFAULT_DEGREE.

    Returns:
        float: Estimated value at given point x0.

    Raises:
        ValueError: If the degree of local polynomial regression is negative or not an integer.
    """
    if p < 0 or not isinstance(p, int):
        raise ValueError("The degree of local polynomial regression p must be a non-negative integer.")
    X = np.array([[((x[i] - x0) ** j) for j in range(p + 1)] for i in range(len(x))])
    W = np.diag(k((x - x0) / h))
    l = np.dot(np.insert(np.zeros(p), 0, 1), np.linalg.pinv(X.T.dot(W).dot(X)).dot(X.T).dot(W))
    return sum(l * y)


import functools
import warnings

import numpy as np

from nopapy.utils.setting import DEFAULT_BANDWIDTH, DEFAULT_KERNEL


def check_estimate(func):
    """ Decorator to check the x0 input of the function. """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        x, y, x0, *args_ = args
        if not isinstance(x0, (int, float, np.int32)):
            raise TypeError(f"{func.__name__}() expects x0 to be an integer or float or numpy.int32.")
        return func(*args, **kwargs)

    return wrapper


def check_regression(func):
    """ Decorator to check the xs0 input of the function. """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        x, y, xs0, *args_ = args
        if not isinstance(xs0, (np.ndarray, list)):
            corr_func = func.__name__.split("Regression")[0] + "Estimate()"
            warn_msg = f"{func.__name__}() expects xs0 to be np.ndarray or list. " \
                       f"Your code won't make any errors, " \
                       f"but we still recommend that you use {corr_func} instead. "
            warnings.warn(warn_msg)

        elif len(xs0) == 1:
            corr_func = func.__name__.split("Regression")[0] + "Estimate()"
            _ = xs0[0]
            warn_msg = f"{func.__name__}() expects xs0 to be a vector, not a scalar. " \
                       f"You may have passed an incorrect value for parameter xs0, " \
                       f"If you want to estimate the dependent variable of a point, " \
                       f"use {corr_func} instead of {func.__name__}, " \
                       f"and use 'x0={_}' instead of 'x0=[{_}]'. "
            warnings.warn(warn_msg)

        return func(*args, **kwargs)

    return wrapper


def check_input(func):
    """ Decorator to check the inputs """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        x, y, *args_ = args
        if not isinstance(x, np.ndarray) or not isinstance(y, np.ndarray):
            raise TypeError(f"{func.__name__}() expects x and y to be np.ndarray.")
        if not x.shape == y.shape:
            raise ValueError(f"{func.__name__}() expects x and y to have the same shape.")
        if not isinstance(kwargs.get('h', DEFAULT_BANDWIDTH), (float, int)):
            raise ValueError(f"{func.__name__}() expects h to be a float or int.")
        if kwargs.get('h', DEFAULT_BANDWIDTH) <= 0:
            raise ValueError(f"{func.__name__}() expects h to be a positive number.")
        if not callable(kwargs.get('k', DEFAULT_KERNEL)):
            raise TypeError(f"{func.__name__}() expects k to be a function.")
        return func(*args, **kwargs)

    return wrapper


import numpy as np

from nopapy.estimate.estimate import NWEstimate, GMEstimate, LPEstimate
from nopapy.utils.check import check_input, check_regression
from nopapy.utils.setting import DEFAULT_BANDWIDTH, DEFAULT_KERNEL, DEFAULT_DEGREE

__all__ = ["NWRegression", "GMRegression", "LPRegression"]

@check_input
@check_regression
def NWRegression(x: np.array, y: np.array, xs0: np.array, h=DEFAULT_BANDWIDTH, k=DEFAULT_KERNEL) -> np.array:
    """
    Nadaraya-Watson regression.
    Estimate the target values at xs0 using the kernel function k and bandwidth h.

    Parameters:
        x (np.array): Independent variable.
        y (np.array): Dependent variable.
        xs0 (np.array): Points at which to estimate the target values.
        h (float, optional): Bandwidth parameter. Defaults to DEFAULT_BANDWIDTH.
        k (callable, optional): Kernel function. Defaults to DEFAULT_KERNEL.

    Returns:
        np.array: Estimated values at given point xs0.
    """
    return np.array([NWEstimate(x, y, x0, h, k) for x0 in np.atleast_1d(xs0)])


@check_input
@check_regression
def GMRegression(x: np.array, y: np.array, xs0: np.array, h=DEFAULT_BANDWIDTH, k=DEFAULT_KERNEL) -> np.array:
    """
    Gasser-Muller regression.
    Estimate the target values at xs0 using the kernel function k and bandwidth h.

    Parameters:
        x (np.array): Independent variable.
        y (np.array): Dependent variable.
        xs0 (np.array): Points at which to estimate the target values.
        h (float, optional): Bandwidth parameter. Defaults to DEFAULT_BANDWIDTH.
        k (callable, optional): Kernel function. Defaults to DEFAULT_KERNEL.

    Returns:
        np.array: Estimated values at given point xs0.
    """
    return np.array([GMEstimate(x, y, x0, h, k) for x0 in np.atleast_1d(xs0)])


@check_input
@check_regression
def LPRegression(x: np.array, y: np.array, xs0: np.array, h=DEFAULT_BANDWIDTH, k=DEFAULT_KERNEL,
                 p=DEFAULT_DEGREE) -> np.array:
    """
    Local polynomial regression.
    Estimate the target values at xs0 using the kernel function k, bandwidth h and degree p.

    Args:
        x (np.array): Independent variable.
        y (np.array): Dependent variable.
        xs0 (np.array): Points at which to estimate the target values.
        h (float, optional): Bandwidth parameter. Defaults to DEFAULT_BANDWIDTH.
        k (callable, optional): Kernel function. Defaults to DEFAULT_KERNEL.
        p (int, optional): The degree of the local polynomial regression. Defaults to DEFAULT_DEGREE.

    Returns:
        np.array: Estimated values at given point xs0.

    Raises:
        ValueError: If the degree of local polynomial regression is negative or not an integer.
    """
    if p < 0 or not isinstance(p, int):
        raise ValueError("The degree of local polynomial regression p must be a non-negative integer.")
    return np.array([LPEstimate(x, y, x0, h, k, p) for x0 in np.atleast_1d(xs0)])


import numpy as np

from nopapy.kernel.custom import I

__all__ = ["GaussianKernel", "EpanechnikovKernel", "tricubeKernel", "boxcarKernel"]

GAUSSIAN_COEFFICIENT = 1 / np.sqrt(2 * np.pi)
EPANECHNIKOV_COEFFICIENT = 0.75
TRICUBE_COEFFICIENT = 70 / 81
BOX_COEFFICIENT = 0.5

def GaussianKernel(x):
    return GAUSSIAN_COEFFICIENT * np.exp(-np.power(x, 2) / 2)


def EpanechnikovKernel(x):
    return EPANECHNIKOV_COEFFICIENT * (1 - np.power(x, 2)) * I(x)


def tricubeKernel(x):
    return TRICUBE_COEFFICIENT * np.power(1 - np.power(np.abs(x), 3), 3) * I(x)


def boxcarKernel(x):
    return BOX_COEFFICIENT * I(x)


from setuptools import setup

def readme():
    with open("README.rst", encoding = "UTF-8") as f:
        return f.read()

setup(
    name = "nopapy",
    version = "1.0.2",
    description = "「NopaPy」is an easy-to-use open-source library for non-parametric statistic.",
    packages = ["nopapy.estimate", "nopapy.kernel", "nopapy.regression", "nopapy.utils"],
    author = "Zhu Yaolin",
    author_email = "540048506@qq.com",
    long_description = readme(),
    include_package_data = True,
    url = "https://github.com/QVQZZZ/NopaPy",
    license = "MIT"
)



import numpy as np
from scipy.integrate import quad

__all__ = ["I", "is_kernel"]

def I(x):
    return (np.abs(x) <= 1).astype(int)

def is_kernel(func, tol=1e-8):
    x = np.linspace(-5, 5, num=1000)
    y = func(x)
    if not np.all(y >= 0):
        return False
    if not np.allclose(y, y[::-1], atol=tol):
        return False
    if not np.isclose(quad(func, -np.inf, np.inf)[0], 1, atol=tol):
        return False
    return True